"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math
from typing import Any, Dict, Optional, Type
import torch
import copy
from torch import nn
from .marginal import DSFMarginal


def _merge_series_time_dims(x: torch.Tensor) -> torch.Tensor:
    """
    Convert a Tensor with dimensions [batch, series, time steps, ...] to one with dimensions [batch, series * time steps, ...]
    """
    assert x.dim() >= 3
    return x.view((x.shape[0], x.shape[1] * x.shape[2]) + x.shape[3:])


def _split_series_time_dims(x: torch.Tensor, target_shape: torch.Size) -> torch.Tensor:
    """
    Convert a Tensor with dimensions [batch, series * time steps, ...] to one with dimensions [batch, series, time steps, ...]
    """
    assert x.dim() + 1 == len(target_shape)
    return x.view(target_shape)


def _simple_linear_projection(input_dim: int, output_dim: int) -> nn.Sequential:
    layers = [nn.Linear(input_dim, output_dim)]
    return nn.Sequential(*layers)


def _easy_mlp(
    input_dim: int,
    hidden_dim: int,
    output_dim: int,
    num_layers: int,
    activation: Type[nn.Module],
) -> nn.Sequential:
    """
    Generate a MLP with the given parameters.
    """
    elayers = [nn.Linear(input_dim, hidden_dim), activation()]
    for _ in range(1, num_layers):
        elayers += [nn.Linear(hidden_dim, hidden_dim), activation()]
    elayers += [nn.Linear(hidden_dim, output_dim)]
    return nn.Sequential(*elayers)


class CopulaDecoder(nn.Module):
    """
    A decoder which forecast using a distribution built from a copula and marginal distributions.
    """

    def __init__(
        self,
        flow_input_dim: int,
        copula_input_dim: int,
        min_u: float = 0.0,
        max_u: float = 1.0,
        skip_sampling_marginal: bool = False,
        attentional_copula: Optional[Dict[str, Any]] = None,
        dsf_marginal: Optional[Dict[str, Any]] = None,
        skip_copula=False,
    ):
        """
        Parameters:
        -----------
        input_dim: int
            The dimension of the encoded representation (upstream data encoder).
        min_u: float, default to 0.0
        max_u: float, default to 1.0
            The values sampled from the copula will be scaled from [0, 1] to [min_u, max_u] before being sent to the marginal.
        skip_sampling_marginal: bool, default to False
            If set to True, then the output from the copula will not be transformed using the marginal during sampling.
            Does not impact the other transformations from observed values to the [0, 1] range.
        trivial_copula: Dict[str, Any], default to None
            If set to a non-None value, uses a TrivialCopula.
            The options sent to the TrivialCopula is content of this dictionary.
        attentional_copula: Dict[str, Any], default to None
            If set to a non-None value, uses a AttentionalCopula.
            The options sent to the AttentionalCopula is content of this dictionary.
        dsf_marginal: Dict[str, Any], default to None
            If set to a non-None value, uses a DSFMarginal.
            The options sent to the DSFMarginal is content of this dictionary.
        """
        super().__init__()

        self.flow_input_dim = flow_input_dim
        self.copula_input_dim = copula_input_dim
        self.min_u = min_u
        self.max_u = max_u
        self.skip_sampling_marginal = skip_sampling_marginal
        self.attentional_copula_args = attentional_copula
        self.dsf_marginal_args = dsf_marginal
        self.skip_copula = skip_copula

        if not self.skip_copula:
            if attentional_copula is not None:
                self.copula = AttentionalCopula(
                    **attentional_copula
                )

        if dsf_marginal is not None:
            self.marginal = DSFMarginal(context_dim=flow_input_dim, **dsf_marginal)

        self.copula_loss = None
        self.marginal_logdet = None

    def create_attentional_copula(self):
        self.skip_copula = False
        if self.attentional_copula_args is not None:
            self.copula = AttentionalCopula(**self.attentional_copula_args)

    def loss(
        self,
        flow_encoded: torch.Tensor,
        copula_encoded: torch.Tensor,
        mask: torch.BoolTensor,
        true_value: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the loss function of the decoder.

        Parameters:
        -----------
        encoded: Tensor [batch, series, time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            This embedding is coming from the encoder, so contains shared information across series and time steps.
        mask: BoolTensor [batch, series, time steps]
            A tensor containing a mask indicating whether a given value was available for the encoder.
            The decoder only forecasts values for which the mask is set to False.
        true_value: Tensor [batch, series, time steps]
            A tensor containing the true value for the values to be forecasted.
            Only the values where the mask is set to False will be considered in the loss function.

        Returns:
        --------
        loss: torch.Tensor [batch]
            The loss function, equal to the negative log likelihood of the distribution.
        """
        B, S, T, E = flow_encoded.shape
        flow_encoded = _merge_series_time_dims(flow_encoded)
        if not self.skip_copula:
            copula_encoded = _merge_series_time_dims(copula_encoded)

        mask = _merge_series_time_dims(mask)
        true_value = _merge_series_time_dims(true_value)

        # Assume that the mask is constant inside the batch (every sample has the same hist-pred split)
        mask = mask[0, :]  # (series * time steps)

        hist_encoded_flow = flow_encoded[:, mask, :]  # [batch, series*(num_hist_timesteps), embedding]
        if not self.skip_copula:
            hist_encoded_copula = copula_encoded[:, mask, :]  # [batch, series*(num_hist_timesteps), embedding]
        pred_encoded_flow = flow_encoded[:, ~mask, :]
        if not self.skip_copula:
            pred_encoded_copula = copula_encoded[:, ~mask, :]
        history_factor = hist_encoded_flow.shape[1] / pred_encoded_flow.shape[1]

        hist_true_x = true_value[:, mask]  # [batch, series*(num_hist_timesteps)]
        pred_true_x = true_value[:, ~mask]

        num_pred_variables = round(T / (history_factor + 1))

        # Transform to [0,1] using the marginals
        hist_true_u = self.marginal.forward_no_logdet(hist_encoded_flow, hist_true_x)
        pred_true_u, marginal_logdet = self.marginal.forward_logdet(pred_encoded_flow, pred_true_x)

        if not self.skip_copula:
            copula_loss = self.copula.loss(
                hist_encoded=hist_encoded_copula,
                hist_true_u=hist_true_u,
                pred_encoded=pred_encoded_copula,
                pred_true_u=pred_true_u,
                num_series=S,
                num_timesteps=num_pred_variables,
            )
        else:
            copula_loss = torch.tensor(0.0).to(mask.device)

        self.copula_loss = copula_loss
        self.marginal_logdet = marginal_logdet

        # Loss = negative log likelihood
        return copula_loss - marginal_logdet

    def sample(
        self,
        num_samples: int,
        flow_encoded: torch.Tensor,
        copula_encoded: torch.Tensor,
        mask: torch.Tensor,
        true_value: torch.Tensor,
    ):
        """
        Generate the given number of samples from the forecasted distribution.

        Parameters:
        -----------
        num_samples: int
            How many samples to generate, must be >= 1.
        encoded: Tensor [batch, series, time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            This embedding is coming from the encoder, so contains shared information across series and time steps.
        mask: BoolTensor [batch, series, time steps]
            A tensor containing a mask indicating whether a given value is masked (available) for the encoder.
            The decoder only forecasts values for which the mask is set to False.
        true_value: Tensor [batch, series, time steps]
            A tensor containing the true value for the values to be forecasted.
            The values where the mask is set to True will be copied as-is in the output.

        Returns:
        --------
        samples: torch.Tensor [batch, series, time steps, samples]
            Samples drawn from the forecasted distribution.
        """
        target_shape = torch.Size((true_value.shape[0], true_value.shape[1], true_value.shape[2], num_samples))

        flow_encoded = _merge_series_time_dims(flow_encoded)
        if not self.skip_copula:
            copula_encoded = _merge_series_time_dims(copula_encoded)

        mask = _merge_series_time_dims(mask)
        true_value = _merge_series_time_dims(true_value)

        # Assume that the mask is constant inside the batch
        mask = mask[0, :]
        hist_true_x = true_value[:, mask]

        hist_encoded_flow = flow_encoded[:, mask, :]
        if not self.skip_copula:
            hist_encoded_copula = copula_encoded[:, mask, :]
        pred_encoded_flow = flow_encoded[:, ~mask, :]
        if not self.skip_copula:
            pred_encoded_copula = copula_encoded[:, ~mask, :]

        # Transform to [0,1] using the marginals
        hist_true_u = self.marginal.forward_no_logdet(hist_encoded_flow, hist_true_x)

        if not self.skip_copula:
            pred_samples = self.copula.sample(
                num_samples=num_samples,
                hist_encoded=hist_encoded_copula,
                hist_true_u=hist_true_u,
                pred_encoded=pred_encoded_copula,
            )
        else:
            num_batches, num_variables, _ = pred_encoded_flow.shape
            pred_samples = torch.rand(num_batches, num_variables, num_samples, device=pred_encoded_flow.device)

        if not self.skip_sampling_marginal:
            # Transform away from [0,1] using the marginals
            # Transform to [min_u, max_u]
            pred_samples = self.min_u + (self.max_u - self.min_u) * pred_samples
            # Transform to the distribution of each token
            # Note that pred_encoded is passed here since it has token, variable information
            # In addition, pred_encoded also has context of the window from attention with other tokens in the window
            pred_samples = self.marginal.inverse(
                pred_encoded_flow,
                pred_samples,
            )

        samples = torch.zeros(
            target_shape[0],
            target_shape[1] * target_shape[2],
            target_shape[3],
            device=flow_encoded.device,
        )
        samples[:, mask, :] = hist_true_x[:, :, None]
        samples[:, ~mask, :] = pred_samples

        return _split_series_time_dims(samples, target_shape)


class AttentionalCopula(nn.Module):
    """
    A non-parametric copula based on attention between the various variables.
    """

    def __init__(
        self,
        input_dim: int,
        attention_heads: int,
        attention_layers: int,
        attention_dim: int,
        mlp_layers: int,
        mlp_dim: int,
        resolution: int = 10,
        dropout: float = 0.1,
        attention_mlp_class: str = "_easy_mlp",
        activation_function: str = "relu",
    ):
        """
        Parameters:
        -----------
        input_dim: int
            Dimension of the encoded representation.
        attention_heads: int
            How many independant heads the attention layer will have. Each head will have its own independant MLP
            to generate the keys and values.
        attention_layers: int
            How many successive attention layers copula will use. Each layer will have its own independant MLPs
            to generate the keys and values.
        attention_dim: int
            The size of the attention layer output.
        mlp_layers: int
            The number of hidden layers in the MLP that produces the keys and values for the attention layer,
            and in the MLP that takes the attention output to generate the distribution parameter.
        mlp_dim: int
            The size of the hidden layers in the MLP that produces the keys and values for the attention layer,
            and in the MLP that takes the attention output to generate the distribution parameter.
        resolution: int, default to 10
            How many bins to pick from when sampling variables.
            Higher values are more precise, but slower to train.
        dropout: float, default to 0.1
            Dropout parameter for the attention.
        """
        super().__init__()

        self.input_dim = input_dim
        self.attention_heads = attention_heads
        self.attention_layers = attention_layers
        self.attention_dim = attention_dim
        self.mlp_layers = mlp_layers
        self.mlp_dim = mlp_dim
        self.resolution = resolution
        self.dropout = dropout

        output_dims = self.attention_heads * self.attention_dim
        # Parameters for the attention layers in the copula
        # For each layer and each head, we have two MLP to create the keys and values
        # After each layer, we transform the embedding using a feed-forward network, consisting of
        # two linear layer with a ReLu in-between both
        # At the very beginning, we have a linear layer to change the embedding to the proper dimensionality
        self.dimension_shifting_layer = nn.Linear(self.input_dim, self.attention_heads * self.attention_dim)

        if activation_function == "relu":
            activation = nn.ReLU
        else:
            raise NotImplementedError("Activation functions other than ReLU are not implemented")

        if attention_mlp_class == "_easy_mlp":
            mlp_args_key = {
                "input_dim": self.input_dim + 1,  # + 1 since the marginals will be concatenated
                "hidden_dim": self.mlp_dim,
                "output_dim": self.attention_dim,
                "num_layers": self.mlp_layers,
                "activation": activation,
            }
            mlp_args_value = {
                "input_dim": self.input_dim + 1,  # + 1 since the marginals will be concatenated
                "hidden_dim": self.mlp_dim,
                "output_dim": self.attention_dim,
                "num_layers": self.mlp_layers,
                "activation": activation,
            }
            mlp_class = _easy_mlp
        elif attention_mlp_class == "_simple_linear_projection":
            mlp_args_key = {
                "input_dim": self.input_dim + 1,  # + 1 since the marginals will be concatenated
                "output_dim": self.attention_dim,
            }
            mlp_args_value = {
                "input_dim": self.input_dim + 1,  # + 1 since the marginals will be concatenated
                "output_dim": self.attention_dim,
            }
            mlp_class = _simple_linear_projection
        else:
            raise NotImplementedError()

        key_attention_heads = self.attention_heads
        value_attention_heads = self.attention_heads

        # one per layer and per head
        # The key and value creators take the input embedding together with the sampled [0,1] value as an input
        self.key_creators = nn.ModuleList(
            [
                nn.ModuleList([mlp_class(**mlp_args_key) for _ in range(key_attention_heads)])
                for _ in range(self.attention_layers)
            ]
        )
        self.value_creators = nn.ModuleList(
            [
                nn.ModuleList([mlp_class(**mlp_args_value) for _ in range(value_attention_heads)])
                for _ in range(self.attention_layers)
            ]
        )

        # one per layer
        self.attention_dropouts = nn.ModuleList([nn.Dropout(self.dropout) for _ in range(self.attention_layers)])
        self.attention_layer_norms = nn.ModuleList(
            [nn.LayerNorm(output_dims, elementwise_affine=False) for _ in range(self.attention_layers)]
        )
        self.feed_forwards = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Linear(output_dims, output_dims),
                    nn.ReLU(),
                    nn.Dropout(self.dropout),
                    nn.Linear(output_dims, output_dims),
                    nn.Dropout(dropout),
                )
                for _ in range(self.attention_layers)
            ]
        )
        self.feed_forward_layer_norms = nn.ModuleList(
            [nn.LayerNorm(output_dims, elementwise_affine=False) for _ in range(self.attention_layers)]
        )

        # Parameter extractor for the categorical distribution
        self.dist_extractors = _easy_mlp(
            input_dim=output_dims,
            hidden_dim=self.mlp_dim,
            output_dim=self.resolution,
            num_layers=self.mlp_layers,
            activation=nn.ReLU,
        )

    def loss(
        self,
        hist_encoded: torch.Tensor,
        hist_true_u: torch.Tensor,
        pred_encoded: torch.Tensor,
        pred_true_u: torch.Tensor,
        num_series: int,
        num_timesteps: int,
    ) -> torch.Tensor:
        """
        Compute the loss function of the copula portion of the decoder.

        Parameters:
        -----------
        hist_encoded: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each series and time step that does not have to be forecasted.
            The series and time steps dimensions are merged.
        hist_true_u: Tensor [batch, series * time steps]
            A tensor containing the true value for the values that do not have to be forecasted, transformed by the marginal distribution into U(0,1) values.
            The series and time steps dimensions are merged.
        pred_encoded: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step that does have to be forecasted.
            The series and time steps dimensions are merged.
        pred_true_u: Tensor [batch, series * time steps]
            A tensor containing the true value for the values to be forecasted, transformed by the marginal distribution into U(0,1) values.
            The series and time steps dimensions are merged.

        Returns:
        --------
        loss: torch.Tensor [batch]
            The loss function, equal to the negative log likelihood of the copula.
        """
        num_batches = pred_encoded.shape[0]
        num_variables = pred_encoded.shape[1]
        num_history = hist_encoded.shape[1]
        device = pred_encoded.device

        assert num_variables == num_series * num_timesteps, (
            "num_variables:"
            + str(num_variables)
            + " but num_series:"
            + str(num_series)
            + " and num_timesteps:"
            + str(num_timesteps)
        )

        permutation = torch.arange(0, num_variables).long()

        # Permute the variables according the random permutation
        pred_encoded = pred_encoded[:, permutation, :]
        pred_true_u = pred_true_u[:, permutation]

        # At the beginning of the attention, we start with the input embedding.
        # Since it does not necessarily have the same dimensions as the hidden layers, apply a linear layer to scale it up.
        att_value = self.dimension_shifting_layer(pred_encoded)

        # # The MLP which generates the keys and values used the encoded embedding + transformed true values.
        # key_value_input_hist = torch.cat([hist_encoded, hist_true_u[:, :, None]], axis=2)
        # key_value_input_pred = torch.cat([pred_encoded, pred_true_u[:, :, None]], axis=2)
        # # key_value_input shape: [bsz, num_history+num_variables, embedding dimension+1]
        # key_value_input = torch.cat([key_value_input_hist, key_value_input_pred], axis=1)

        keys = []
        values = []
        for layer in range(self.attention_layers):
            key_input_hist = torch.cat([hist_encoded, hist_true_u[:, :, None]], axis=2)
            key_input_pred = torch.cat([pred_encoded, pred_true_u[:, :, None]], axis=2)
            key_input = torch.cat([key_input_hist, key_input_pred], axis=1)

            value_input_hist = torch.cat([hist_encoded, hist_true_u[:, :, None]], axis=2)
            value_input_pred = torch.cat([pred_encoded, pred_true_u[:, :, None]], axis=2)
            value_input = torch.cat([value_input_hist, value_input_pred], axis=1)

            # Keys shape in every layer: [bsz, num_attention_heads, num_history+num_variables, attention_dim]
            keys.append(
                torch.cat(
                    [mlp(key_input)[:, None, :, :] for mlp in self.key_creators[layer]],
                    axis=1,
                )
            )
            # Values shape in every layer: [bsz, num_attention_heads, num_history+num_variables, attention_dim]
            values.append(
                torch.cat(
                    [mlp(value_input)[:, None, :, :] for mlp in self.value_creators[layer]],
                    axis=1,
                )
            )

        for layer in range(self.attention_layers):
            # Split the hidden layer into its various heads
            # Basically the Query in the attention
            # Shape: [bsz, num_variables, num_attention_heads, attention_dim]
            att_value_heads = att_value.reshape(
                att_value.shape[0],
                att_value.shape[1],
                self.attention_heads,
                self.attention_dim,
            )

            # Attention layer, for each batch and head:
            # A_vi' = sum_w(softmax_w(sum_i(Q_vi * K_wi) / sqrt(d)) * V_wi')

            # During attention, we will add -float("inf") to pairs of indices where the variable to be forecasted (query)
            # is after the variable that gives information (key), after the random permutation.
            # Doing this prevent information from flowing from later in the permutation to before in the permutation,
            # which cannot happen during inference.
            # tril fill the diagonal and values that are below it, flip rotates it by 180 degrees,
            # leaving only the pairs of indices which represent not yet sampled values.
            # Note float("inf") * 0 is unsafe, so do the multiplication inside the torch.tril()
            # pred/hist_encoded dimensions: number of batches, number of variables, size of embedding per variable
            product_mask = torch.ones(
                num_batches,
                self.attention_heads,
                num_variables,
                num_variables + num_history,
                device=device,
            )
            product_mask = torch.tril(float("inf") * product_mask).flip((2, 3))

            # Perform Attention
            # Einstein sum indices:
            # b: batch number
            # h: attention head number (Note the change in order for att_value_heads)
            # v: variable we want to predict
            # w: variable we want to get information from (history or prediction)
            # i: embedding dimension of the keys and queries (self.attention_dim)
            # Output shape: [bsz, num_attention_heads, num_variables, num_history+num_variables]
            product_base = torch.einsum("bvhi,bhwi->bhvw", att_value_heads, keys[layer])

            # Adding -inf shunts the attention to zero, for any variable that has not "yet" been predicted,
            # aka: are in the future according to the permutation.
            product = product_base - product_mask
            product = self.attention_dim ** (-0.5) * product
            weights = nn.functional.softmax(product, dim=-1)

            # Einstein sum indices:
            # b: batch number
            # h: attention head number (Note the change in order for the result)
            # v: variable we want to predict
            # w: variable we want to get information from (history or prediction)
            # j: embedding dimension of the values (self.attention_dim)
            # Output shape: [bsz, num_variables, num_attention_heads, attention_dim]
            att = torch.einsum("bhvw,bhwj->bvhj", weights, values[layer])

            # Merge back the various heads to allow the feed forwards module to share information between heads
            # Shape: [b, v, h*j]
            att_merged_heads = att.reshape(att.shape[0], att.shape[1], att.shape[2] * att.shape[3])
            # print("Q:", att_value_heads.shape, "K:", keys[layer].shape, "V:", values[layer].shape, "Mask:", product_mask.shape)
            # print("Attn:", att_merged_heads.shape, "Attn Scores", weights.shape)

            # Compute and add dropout
            att_merged_heads = self.attention_dropouts[layer](att_merged_heads)

            # att_value = att_value + att_merged_heads
            # Layernorm
            att_value = att_value + att_merged_heads
            att_value = self.attention_layer_norms[layer](att_value)

            # Add the contribution of the feed-forward layer, mixing up the heads
            att_feed_forward = self.feed_forwards[layer](att_value)
            att_value = att_value + att_feed_forward
            att_value = self.feed_forward_layer_norms[layer](att_value)

        # Assign each observed U(0,1) value to a bin. The clip is to avoid issues with numerical inaccuracies.
        # shape: [b, variables*timesteps - 1]
        target = torch.clip(
            torch.floor(pred_true_u[:, 1:] * self.resolution).long(),
            min=0,
            max=self.resolution - 1,
        )

        # Final shape of att_value would be: [bsz, num_variables, num_attention_heads*attention_dim]
        # Compute the (un-normalized) logarithm likelihood of the conditional distribution.
        # Note: This section could instead call a specialized module to allow for easier customization.
        # Get conditional distributions over bins for all variables but the first one.
        # The first one is considered to always be U(0,1), which has a constant logarithm likelihood of 0.
        logits = self.dist_extractors(att_value)[:, 1:, :]  # shape: [b, variables*timesteps - 1, self.resolution]

        # We multiply the probability by self.resolution to get the PDF of the continuous-by-part distribution.
        # prob = self.resolution * softmax(logits, dim=2)
        # logprob = log(self.resolution) + log_softmax(logits, dim=2)
        # Basically softmax of the logits, but the softmax is scaled according to the resolution instead of being in [0,1]
        logprob = math.log(self.resolution) + nn.functional.log_softmax(logits, dim=2)
        # For each batch + variable pair, we want the value of the logits associated with its true value (target):
        # logprob[batch,variable] = logits[batch,variable,target[batch,variable]]
        # Since gather wants the same number of dimensions for both tensors, add and remove a dummy third dimension.
        logprob = torch.gather(logprob, dim=2, index=target[:, :, None])[:, :, 0]

        return -logprob.sum(axis=1)  # Only keep the batch dimension

    def sample(
        self,
        num_samples: int,
        hist_encoded: torch.Tensor,
        hist_true_u: torch.Tensor,
        pred_encoded: torch.Tensor,
    ) -> torch.Tensor:
        """
        Generate the given number of samples from the forecasted copula.

        Parameters:
        -----------
        num_samples: int
            How many samples to generate, must be >= 1.
        hist_encoded: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step that does not have to be forecasted.
            The series and time steps dimensions are merged.
        hist_true_u: Tensor [batch, series * time steps]
            A tensor containing the true value for the values that do not have to be forecasted, transformed by the marginal distribution into U(0,1) values.
            The series and time steps dimensions are merged.
        pred_encoded: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step that does have to be forecasted.
            The series and time steps dimensions are merged.

        Returns:
        --------
        samples: torch.Tensor [batch, series * time steps, samples]
            Samples drawn from the forecasted copula, thus in the [0, 1] range.
            The series and time steps dimensions are merged.
        """
        num_batches = pred_encoded.shape[0]
        num_variables = pred_encoded.shape[1]
        num_history = hist_encoded.shape[1]
        device = pred_encoded.device

        permutation = torch.arange(0, num_variables).long()
        permutations = torch.stack([permutation for _ in range(num_samples)])

        key_value_input_hist = torch.cat([hist_encoded, hist_true_u[:, :, None]], axis=2)

        # The MLP which generates the keys and values used the encoded embedding + transformed true values.
        # Keys and values shape in every layer FOR HISTORY: [bsz, num_attention_heads, num_history, attention_dim]
        keys_hist = [
            torch.cat(
                [mlp(key_value_input_hist)[:, None, :, :] for mlp in self.key_creators[layer]],
                axis=1,
            )
            for layer in range(self.attention_layers)
        ]
        values_hist = [
            torch.cat(
                [mlp(key_value_input_hist)[:, None, :, :] for mlp in self.value_creators[layer]],
                axis=1,
            )
            for layer in range(self.attention_layers)
        ]

        # We will store the keys and values from the sampled variables as we do the sampling
        # We have an additional dimension num_samples here
        samples = torch.zeros(num_batches, num_variables, num_samples).to(device)
        keys_samples = [
            torch.zeros(
                num_batches,
                num_samples,
                self.attention_heads,
                num_variables,
                self.attention_dim,
                device=device,
            )
            for _ in range(self.attention_layers)
        ]
        values_samples = [
            torch.zeros(
                num_batches,
                num_samples,
                self.attention_heads,
                num_variables,
                self.attention_dim,
                device=device,
            )
            for _ in range(self.attention_layers)
        ]

        # Importantly, at every step (for every variable), the attention outputs a vector of size [bsz, num_samples]
        # We sample the copula one variable at a time, following the order from the drawn permutation (for each sample)
        # Sampling of the ith variable is done for all samples
        for i in range(num_variables):
            # Vector containing which variable we sample at this step of the copula. It is different for each sample.
            # Shape: [num_samples]
            p = permutations[:, i]
            # Note that second dimension here no longer represent the variables (as in the loss method), but the samples.
            # Shape: [num_batches, num_samples, embedding dimension]
            current_pred_encoded = pred_encoded[:, p, :]

            if i == 0:
                # By construction, the first variable to be sampled is always sampled according to a Uniform(0,1).
                current_samples = torch.rand(num_batches, num_samples, device=device)
            else:
                att_value = self.dimension_shifting_layer(current_pred_encoded)

                for layer in range(self.attention_layers):
                    # Split the hidden layer into its various heads
                    # Basically the Query in the attention
                    # Shape: [bsz, num_samples, num_attention_heads, attention_dim]
                    att_value_heads = att_value.reshape(
                        att_value.shape[0],
                        att_value.shape[1],
                        self.attention_heads,
                        self.attention_dim,
                    )

                    keys_hist_current_layer = keys_hist[layer]
                    keys_samples_current_layer = keys_samples[layer][:, :, :, 0:i, :]
                    values_hist_current_layer = values_hist[layer]
                    values_samples_current_layer = values_samples[layer][:, :, :, 0:i, :]

                    # Calculate attention weights
                    # Einstein sum indices:
                    # b: batch number
                    # n: sample number of the current variable (treated as a separate variable)
                    # h: attention head number
                    # w: variable we want to get information from (history or prediction)
                    # i: embedding dimension of the keys and queries (self.input_dim)
                    product_hist = torch.einsum("bnhi,bhwi->bnhw", att_value_heads, keys_hist_current_layer)
                    # keys_samples is full of zero starting at i of the 4th dimension (w)
                    # Shape of product_samples: [bsz, num_samples, num_attention_heads, i]
                    # i since only i variables can be seen
                    product_samples = torch.einsum("bnhi,bnhwi->bnhw", att_value_heads, keys_samples_current_layer)
                    # # NOTE: product_samples could also be computed as follows
                    # b, n, h, d = att_value_heads.shape
                    # b, n, h, w, d = keys_samples_current_layer.shape
                    # product_samples = torch.einsum(
                    #     "bhi,bhwi->bhw", att_value_heads.reshape(b*n, h, d), keys_samples_current_layer.reshape(b*n, h, w, d)
                    # ).reshape(b, n, h, w)

                    # Combine the attention from the history and from the previous samples.
                    # Product is of shape [bsz, num_samples, num_attention_heads, num_history+i]
                    product = torch.cat([product_hist, product_samples], axis=3)
                    product = self.attention_dim ** (-0.5) * product

                    weights = nn.functional.softmax(product, dim=3)
                    weights_hist = weights[:, :, :, :num_history]
                    weights_samples = weights[:, :, :, num_history:]

                    # Get attention representation using weights (for conditional distribution)
                    # Einstein sum indices:
                    # b: batch number
                    # n: sample number of the current variable (treated as a separate variable)
                    # h: attention head number
                    # w: variable we want to get information from (history or prediction)
                    # j: embedding dimension of the values (self.hid_dim)
                    # att_hist is of shape [bsz, num_samples, num_attention_heads, hid_dim]
                    att_hist = torch.einsum("bnhw,bhwj->bnhj", weights_hist, values_hist_current_layer)
                    # att_hist is of shape [bsz, num_samples, num_attention_heads, hid_dim]
                    att_samples = torch.einsum(
                        "bnhw,bnhwj->bnhj",
                        weights_samples,
                        values_samples_current_layer,
                    )  # i >= 1
                    # # NOTE: att_samples could also be computed as follows
                    # b, n, h, w = weights_samples.shape
                    # b, n, h, w, j = values_samples_current_layer.shape
                    # att_samples = torch.einsum(
                    #     "bhw,bhwj->bhj", weights_samples.reshape(b*n, h, w), values_samples_current_layer.reshape(b*n, h, w, j)
                    # ).reshape(b, n, h, j)

                    att = att_hist + att_samples

                    # Merge back the various heads to allow the feed forwards module to share information between heads
                    att_merged_heads = att.reshape(att.shape[0], att.shape[1], att.shape[2] * att.shape[3])

                    att_merged_heads = self.attention_dropouts[layer](att_merged_heads)
                    att_value = att_value + att_merged_heads
                    att_value = self.attention_layer_norms[layer](att_value)
                    att_feed_forward = self.feed_forwards[layer](att_value)
                    att_value = att_value + att_feed_forward
                    att_value = self.feed_forward_layer_norms[layer](att_value)

                # Below code is executed once all the layers of attention are complete
                # att_value is of shape [bsz, num_samples, attn_heads*attn_dim]
                # Get the output distribution parameters
                logits = self.dist_extractors(att_value).reshape(num_batches * num_samples, self.resolution)

                # Select a single variable in {0, 1, 2, ..., self.resolution-1} according to the probabilities from the softmax
                # Why not the variable corresponding to the maximum value?
                current_samples = torch.multinomial(input=torch.softmax(logits, dim=1), num_samples=1)
                # Each point in the same bucket is equiprobable, and we used a floor function in the training
                current_samples = current_samples + torch.rand(*current_samples.shape).to(device)
                # Normalize to a variable in the [0, 1) range
                current_samples /= self.resolution
                current_samples = current_samples.reshape(num_batches, num_samples)

            # Compute the key and value associated with the newly sampled variable, for the attention of the next ones.
            # Shape: [num_batches, num_samples, embedding dimension+1]
            key_value_input = torch.cat([current_pred_encoded, current_samples[:, :, None]], axis=-1)
            for layer in range(self.attention_layers):
                # Shape: [num_batches, num_samples, num_attention_heads, attn_dim]
                new_keys = torch.cat(
                    [k(key_value_input)[:, :, None, :] for k in self.key_creators[layer]],
                    axis=2,
                )
                # Shape: [num_batches, num_samples, num_attention_heads, attn_dim]
                new_values = torch.cat(
                    [v(key_value_input)[:, :, None, :] for v in self.value_creators[layer]],
                    axis=2,
                )
                # Store the computed keys and values for the ith variable in the ith slot
                keys_samples[layer][:, :, :, i, :] = new_keys
                values_samples[layer][:, :, :, i, :] = new_values

            # p is of length num_samples, and each value indicates the variable sampled at this i'th timestep
            # Store the samples at the respective positions indicated by `p` for each sample
            # Collate the results, reversing the effect of the permutation
            # By using two lists of equal lengths, the resulting slice will be 2d, not 3d.
            samples[:, p, range(num_samples)] = current_samples

        return samples



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
from typing import Optional, Union

import torch
import os
import torch.nn as nn
from torch.optim import Adam, RMSprop
from torch.utils.data import DataLoader

from gluonts.core.component import validated
from tactis.gluon.utils import (
    save_checkpoint,
    set_seed,
)
from pts import Trainer

from gluonts.dataset.loader import ValidationDataLoader


class TACTISTrainer(Trainer):
    @validated()
    def __init__(
        self,
        epochs: int = 100,
        epochs_phase_1 = None,
        epochs_phase_2 = None,
        batch_size: int = 32,
        training_num_batches_per_epoch: int = 50,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-6,
        clip_gradient: Optional[float] = None,
        device: Optional[Union[torch.device, str]] = None,
        log_subparams_every=-1,
        checkpoint_dir=None,
        seed=42,
        load_checkpoint=None,
        early_stopping_epochs=-1,
        do_not_restrict_time=False,
        **kwargs,
    ) -> None:
        """
        When `epochs_phase_1` and `epochs_phase_2` are not specified, training is done for a total of `epochs` epochs, and `early_stopping_epochs` is used to switch from phase 1 to phase 2.
        When `epochs_phase_1` and `epochs_phase_2` are specified, `early_stopping_epochs` is ignored, and phase 1 and phase 2 are trained for `epochs_phase_1` and `epochs_phase_2` epochs respectively.
        """
        self.epochs = epochs
        self.epochs_phase_1 = epochs_phase_1
        self.epochs_phase_2 = epochs_phase_2
        self.batch_size = batch_size
        self.training_num_batches_per_epoch = training_num_batches_per_epoch
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.clip_gradient = clip_gradient
        self.device = device
        self.log_subparams_every = log_subparams_every
        self.checkpoint_dir = checkpoint_dir
        self.seed = seed
        self.load_checkpoint = load_checkpoint

        self.early_stopping_epochs = early_stopping_epochs
        self.do_not_restrict_time = do_not_restrict_time

        if self.checkpoint_dir is None:
            print("Checkpoints will not be saved")
        else:
            print("Checkpoints will be saved at", self.checkpoint_dir)
            os.makedirs(self.checkpoint_dir, exist_ok=True)

    def set_load_checkpoint(self, load_checkpoint):
        self.load_checkpoint = load_checkpoint

    @staticmethod
    def disable_grads(net, disable_grads):
        total_enabled = 0
        total_disabled = 0
        total_grads = 0
        for j, p in enumerate(net.model.named_parameters()):
            for param in disable_grads:
                if param in p[0]:
                    print("Disabling gradient on", p[0])
                    p[1].requires_grad = False
            if p[1].requires_grad is True:
                total_enabled += 1
            if p[1].requires_grad is False:
                total_disabled += 1
            total_grads += 1

        print("Grads to disable:", disable_grads)
        print("Total disabled grads:", total_disabled, "/", total_grads)
        print("Total enabled grads:", total_enabled, "/", total_grads)

    def __call__(
        self,
        net: nn.Module,
        train_iter: DataLoader,
        validation_iter_args=None,
        optimizer: str = "adam",
    ) -> None:
        set_seed(self.seed)

        # A list of parameters related to each component for reference
        params_input_encoder_flow = [
            "flow_series_encoder",
            "flow_time_encoding",
            "flow_input_encoder",
        ]
        params_encoder_flow = ["flow_encoder"]
        params_decoder_flow = ["decoder.marginal"]

        disable_grads = []

        if self.load_checkpoint:
            set_seed(self.seed)
            assert os.path.isfile(self.load_checkpoint), "Checkpoint " + self.load_checkpoint + "is invalid"
            print("Loading from checkpoint", self.load_checkpoint)
            ckpt = torch.load(self.load_checkpoint, map_location=self.device)
            if ckpt["stage"] == 1:
                print("Loaded checkpoint has stage 1")
                net.model.set_stage(1)
                flow_loss_weight = 1.0
                copula_loss_weight = 0.0
                print("Loading model...")
                net.load_state_dict(ckpt["model"])
                print("Loaded model")
                if optimizer == "rmsprop":
                    optim = RMSprop(
                        net.parameters(),
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )
                elif optimizer == "adam":
                    optim = Adam(
                        net.parameters(),
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )
                print("Loading optim...")
                optim.load_state_dict(ckpt["optim"])
                print("Loaded optim...")
            else:
                print("Loaded checkpoint has stage 2")
                net.model.set_stage(2)
                net.model.initialize_stage2()
                net.to(self.device)

                # Create parameter groups for the specific parameters
                parameter_names_to_optimize = [
                    "copula_series_encoder",
                    "copula_time_encoding",
                    "copula_input_encoder",
                    "copula_encoder",
                    "decoder.copula",
                ]
                params_to_optimize_in_stage2 = []
                for name, param in net.model.named_parameters():
                    if any(pname in name for pname in parameter_names_to_optimize):
                        params_to_optimize_in_stage2.append(param)

                flow_loss_weight = 0.0
                copula_loss_weight = 1.0
                print("Loading model...")
                net.load_state_dict(ckpt["model"])
                print("Loaded model")
                if optimizer == "rmsprop":
                    optim = RMSprop(
                        params_to_optimize_in_stage2,
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )
                elif optimizer == "adam":
                    optim = Adam(
                        params_to_optimize_in_stage2,
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )
                print("Loading optim...")
                optim.load_state_dict(ckpt["optim"])
                print("Loaded optim...")

                disable_grads.extend(params_decoder_flow)
                disable_grads.extend(params_input_encoder_flow)
                disable_grads.extend(params_encoder_flow)
                TACTISTrainer.disable_grads(net, disable_grads)

            start_epoch = ckpt["epoch"] + 1
            print("Start epoch set to", start_epoch)
            best_val_loss_unweighted = ckpt["best_val_loss_unweighted"]
            best_epoch = ckpt["best_epoch"]
            total_training_only_time = ckpt["total_training_only_time"]
        else:
            net.model.set_stage(1)
            flow_loss_weight = 1.0
            copula_loss_weight = 0.0
            if optimizer == "rmsprop":
                optim = RMSprop(
                    net.parameters(),
                    lr=self.learning_rate,
                    weight_decay=self.weight_decay,
                )
            elif optimizer == "adam":
                optim = Adam(
                    net.parameters(),
                    lr=self.learning_rate,
                    weight_decay=self.weight_decay,
                )

        step = -1
        best_epoch = -1

        best_val_loss_unweighted = None

        start_epoch = 0
        epoch_no = start_epoch
        batch_size = self.batch_size
        total_training_only_time = 0.0

        ## Log the start and total epochs
        print("Start epoch:", start_epoch)
        print("Total epochs:", self.epochs)
        print("Epochs:", self.epochs)

        print("Flow loss weight:", flow_loss_weight)
        print("Copula loss weight:", copula_loss_weight)

        ## Iterator for the epochs
        switch_to_stage_2 = False
        if self.epochs_phase_1 and self.epochs_phase_2:
            total_epochs = self.epochs_phase_1 + self.epochs_phase_2
        else:
            total_epochs = self.epochs
        while epoch_no < total_epochs:
            # Stage switching
            if switch_to_stage_2:
                net.model.set_stage(2)
                net.model.initialize_stage2()
                net.to(self.device)

                # Create parameter groups for the specific parameters
                parameter_names_to_optimize = [
                    "copula_series_encoder",
                    "copula_time_encoding",
                    "copula_input_encoder",
                    "copula_encoder",
                    "decoder.copula",
                ]
                params_to_optimize_in_stage2 = []
                for name, param in net.model.named_parameters():
                    if any(pname in name for pname in parameter_names_to_optimize):
                        params_to_optimize_in_stage2.append(param)

                flow_loss_weight = 0.0
                copula_loss_weight = 1.0
                if optimizer == "rmsprop":
                    optim = RMSprop(
                        params_to_optimize_in_stage2,
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )
                elif optimizer == "adam":
                    optim = Adam(
                        params_to_optimize_in_stage2,
                        lr=self.learning_rate,
                        weight_decay=self.weight_decay,
                    )

                disable_grads.extend(params_decoder_flow)
                disable_grads.extend(params_input_encoder_flow)
                disable_grads.extend(params_encoder_flow)

                TACTISTrainer.disable_grads(net, disable_grads)

                epochs_since_best_epoch = 0
                total_training_only_time = 0.0
                switch_to_stage_2 = False

            net.train()

            print("Epoch:", epoch_no, "/", self.epochs)
            print("Training...")
            set_seed(epoch_no + self.seed)
            cumm_epoch_loss = 0.0
            cumm_epoch_loss_unweighted = 0.0
            cumm_epoch_loss_unnormalized = 0.0
            cumm_marginal_logdet = 0.0  # Normalized
            cumm_copula_loss = 0.0  # Normalized

            print("Total number of training batches:", self.training_num_batches_per_epoch)

            # training loop
            training_num_windows_seen = 0
            training_epoch_start_time = time.time()
            for batch_no, data_entry in enumerate(train_iter, start=0):
                step = epoch_no * self.training_num_batches_per_epoch + batch_no
                if self.log_subparams_every != -1 and step % self.log_subparams_every == 0:
                    print("Iter:", batch_no, "/", self.training_num_batches_per_epoch)

                optim.zero_grad()

                inputs = [v.to(self.device)[:batch_size] for v in data_entry.values()]
                training_num_windows_seen += len(inputs[0])

                # Forward pass
                _ = net(*inputs)
                # Loss computation
                marginal_logdet, copula_loss = (
                    net.model.marginal_logdet,
                    net.model.copula_loss,
                )
                loss = copula_loss_weight * copula_loss - flow_loss_weight * marginal_logdet
                loss_avg = loss.mean()
                # Backward pass
                loss_avg.backward()
                if self.clip_gradient is not None:
                    nn.utils.clip_grad_norm_(net.parameters(), self.clip_gradient)
                optim.step()

                cumm_epoch_loss += loss.sum()
                cumm_epoch_loss_unweighted += torch.sum(copula_loss - marginal_logdet).item()
                cumm_epoch_loss_unnormalized += torch.sum(
                    net.model.unnormalized_copula_loss - net.model.marginal_logdet
                ).item()
                cumm_marginal_logdet += -torch.sum(marginal_logdet).item()
                cumm_copula_loss += torch.sum(copula_loss).item()

                if self.training_num_batches_per_epoch == batch_no:
                    break

            # Accumulate training time
            training_epoch_end_time = time.time()
            total_training_only_time += training_epoch_end_time - training_epoch_start_time

            avg_epoch_loss = cumm_epoch_loss / training_num_windows_seen
            print("Epoch:", epoch_no, "Average training loss:", avg_epoch_loss.item())

            ####### VALIDATION #########
            ####### VALIDATION #########
            ####### VALIDATION #########
            print("Validation...")
            if validation_iter_args is not None:
                print("Creating a validation dataloader with a batch size of", batch_size)
                validation_iter = ValidationDataLoader(**validation_iter_args, batch_size=batch_size)
            validation_length = None
            net.eval()
            cumm_epoch_loss_val = 0.0
            cumm_epoch_loss_val_unweighted = 0.0
            cumm_epoch_loss_val_unnormalized = 0.0
            cumm_marginal_loss_val = 0.0
            cumm_copula_loss_val = 0.0
            set_seed(epoch_no + self.seed)

            validation_num_windows_seen = 0
            validation_num_batches_per_epoch = len(validation_iter)
            for batch_no, data_entry in enumerate(validation_iter, start=0):
                if self.log_subparams_every != -1 and step % self.log_subparams_every == 0:
                    print("Iter:", batch_no, "/", validation_num_batches_per_epoch)
                inputs = [v.to(self.device) for v in data_entry.values()]
                validation_num_windows_seen += len(inputs[0])
                if validation_length:
                    inputs = [v[:, :validation_length] for v in inputs]
                with torch.no_grad():
                    _ = net(*inputs)

                    marginal_logdet, copula_loss = (
                        net.model.marginal_logdet,
                        net.model.copula_loss,
                    )
                    loss = copula_loss_weight * copula_loss - flow_loss_weight * marginal_logdet
                    loss_avg = loss.mean()

                cumm_epoch_loss_val += loss.sum()
                cumm_epoch_loss_val_unweighted += torch.sum(copula_loss - marginal_logdet).item()
                cumm_epoch_loss_val_unnormalized += torch.sum(
                    net.model.unnormalized_copula_loss - net.model.marginal_logdet
                ).item()
                cumm_marginal_loss_val += -torch.sum(marginal_logdet).item()
                cumm_copula_loss_val += torch.sum(copula_loss).item()

            avg_epoch_loss_val = cumm_epoch_loss_val / validation_num_windows_seen
            avg_epoch_loss_val_unweighted = cumm_epoch_loss_val_unweighted / validation_num_windows_seen

            print("Epoch:", epoch_no, "Average validation loss:", avg_epoch_loss_val.item())

            if best_val_loss_unweighted == None or avg_epoch_loss_val_unweighted < best_val_loss_unweighted:
                best_val_loss_unweighted = avg_epoch_loss_val_unweighted
                best_epoch = epoch_no
                epochs_since_best_epoch = 0
            else:
                epochs_since_best_epoch += 1

            print("Epochs since best epoch:", epochs_since_best_epoch)

            # If in stage 2, stop. If in stage 1, switch to stage 2 (taken care of in the )
            if self.epochs_phase_1 is None and self.epochs_phase_2 is None and self.early_stopping_epochs != -1 and epochs_since_best_epoch == self.early_stopping_epochs:
                if net.model.stage == 2:
                    print("Stopping criterion reached for stage 2. Stopping training.")
                    break
                else:
                    print("Stopping criterion reached for stage 1. Shifting to stage 2.")
                    switch_to_stage_2 = True

            ####### VALIDATION #########
            ####### VALIDATION #########
            ####### VALIDATION #########

            if self.checkpoint_dir:
                state_dict = {
                    "model": net.state_dict(),
                    "epoch": epoch_no,
                    "step": step,
                    "best_val_loss_unweighted": best_val_loss_unweighted,
                    "best_epoch": best_epoch,
                    "stage": net.model.stage,
                    "total_training_only_time": total_training_only_time,
                }
                state_dict["optim"] = optim.state_dict()
                filename = "last.pth.tar"
                save_checkpoint(state_dict, self.checkpoint_dir, filename=filename)
                print(
                    "Checkpoint of epoch",
                    epoch_no,
                    "saved at",
                    os.path.join(self.checkpoint_dir, filename),
                )

                if epochs_since_best_epoch == 0:
                    filename = "best.pth.tar"
                    save_checkpoint(state_dict, self.checkpoint_dir, filename=filename)
                    print(
                        "Checkpoint of epoch",
                        epoch_no,
                        "saved at",
                        os.path.join(self.checkpoint_dir, filename),
                    )

            # Check if epochs criterion is reached
            if net.model.stage == 1 and self.epochs_phase_1 and epoch_no == self.epochs_phase_1 - 1:
                switch_to_stage_2 = True
            elif net.model.stage == 2 and self.epochs_phase_2 and epoch_no == self.epochs_phase_2 - 1:
                break

            # Check if the training time criterion is reached
            if not self.do_not_restrict_time and total_training_only_time >= 129600:
                if net.model.stage == 2:
                    print("Time limit reached for stage 2. Stopping training.")
                    break
                else:
                    print("Time limit reached for stage 1. Shifting to stage 2.")
                    switch_to_stage_2 = True

            # Increase epoch_no
            epoch_no += 1

        return best_val_loss_unweighted

    def validate(
        self,
        net: nn.Module,
        validation_iter_args=None,
    ):
        if self.load_checkpoint:
            assert os.path.isfile(self.load_checkpoint), "Checkpoint " + self.load_checkpoint + "is invalid"
            print("Loading from checkpoint", self.load_checkpoint)
            ckpt = torch.load(self.load_checkpoint, map_location=self.device)
            net.load_state_dict(ckpt["model"])

        # Note that the NLL you obtain from this function is only comparable to TACTiS-style architecture
        # To make it comparable to non-TACTiS style architectures, you need to denormalize the loss of every sample/batch which is not currently implemented
        print("Validation...")
        batch_size = self.batch_size

        if validation_iter_args is not None:
            print("Creating a validation dataloader with a batch size of", batch_size)
            validation_iter = ValidationDataLoader(**validation_iter_args, batch_size=batch_size)
        validation_length = None
        net.eval()
        cumm_epoch_loss_val = 0.0

        validation_num_windows_seen = 0
        for batch_no, data_entry in enumerate(validation_iter, start=0):
            inputs = [v.to(self.device) for v in data_entry.values()]
            validation_num_windows_seen += len(inputs[0])
            if validation_length:
                inputs = [v[:, :validation_length] for v in inputs]
            with torch.no_grad():
                _ = net(*inputs)

                marginal_logdet, copula_loss = (
                    net.model.marginal_logdet,
                    net.model.copula_loss,
                )
                loss = copula_loss - marginal_logdet

            cumm_epoch_loss_val += loss.sum()

        avg_epoch_loss_val = cumm_epoch_loss_val / validation_num_windows_seen

        return avg_epoch_loss_val



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import torch
from torch import nn
from typing import Tuple
from .flow import DeepSigmoidFlow


class DSFMarginal(nn.Module):
    """
    Compute the marginals using a Deep Sigmoid Flow conditioned using a MLP.
    The conditioning MLP uses the embedding from the encoder as its input.
    """

    def __init__(
        self,
        context_dim: int,
        mlp_layers: int,
        mlp_dim: int,
        flow_layers: int,
        flow_hid_dim: int,
    ):
        """
        Parameters:
        -----------
        context_dim: int
            Size of the context (embedding created by the encoder) that will be sent to the conditioner.
        mlp_layers: int
            Number of layers for the conditioner MLP.
        mlp_dim: int
            Dimension of the hidden layers of the conditioner MLP.
        flow_layers: int
            Number of layers for the Dense Sigmoid Flow.
        flow_hid_dim: int
            Dimension of the hidden layers of the Dense Sigmoid Flow.
        """
        super().__init__()

        self.context_dim = context_dim
        self.mlp_layers = mlp_layers
        self.mlp_dim = mlp_dim
        self.flow_layers = flow_layers
        self.flow_hid_dim = flow_hid_dim

        self.marginal_flow = DeepSigmoidFlow(n_layers=self.flow_layers, hidden_dim=self.flow_hid_dim)

        elayers = [nn.Linear(self.context_dim, self.mlp_dim), nn.ReLU()]
        for _ in range(1, self.mlp_layers):
            elayers += [nn.Linear(self.mlp_dim, self.mlp_dim), nn.ReLU()]
        elayers += [nn.Linear(self.mlp_dim, self.marginal_flow.total_params_length)]
        self.marginal_conditioner = nn.Sequential(*elayers)

    def forward_logdet(self, context: torch.Tensor, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the cumulative density function of a marginal conditioned using the given context, for the given value of x.
        Also returns the logarithm of the derivative of this transformation.

        Parameters:
        -----------
        context: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            The series and time steps dimensions are merged.
        x: Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            A tensor containing the value to be transformed using the CDF.
            The series and time steps dimensions are merged.
            If a third dimension is present, then the context is considered to be constant across this dimension.
        Returns:
        --------
        u: torch.Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            The CDF at the given point, a value between 0 and 1.
            The series and time steps dimensions are merged.
            The shape of the output is the same as the shape of x.
        logdet: torch.Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            The logarithm of the derivative of the transformation.
            The series and time steps dimensions are merged.
            The shape of the output is the same as the shape of x.
        """
        marginal_params = self.marginal_conditioner(context)
        # If x has both a variable and a sample dimension, then add a singleton dimension to marginal_params to have the correct shape
        if marginal_params.dim() == x.dim():
            marginal_params = marginal_params[:, :, None, :]

        return self.marginal_flow.forward(marginal_params, x)

    def forward_no_logdet(self, context: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the cumulative density function of a marginal conditioned using the given context, for the given value of x.

        Parameters:
        -----------
        context: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            The series and time steps dimensions are merged.
        x: Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            A tensor containing the value to be transformed using the CDF.
            The series and time steps dimensions are merged.
            If a third dimension is present, then the context is considered to be constant across this dimension.
        Returns:
        --------
        u: torch.Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            The CDF at the given point, a value between 0 and 1.
            The series and time steps dimensions are merged.
            The shape of the output is the same as the shape of x.
        """
        marginal_params = self.marginal_conditioner(
            context
        )  # [batch, series*timesteps, self.marginal_flow.total_params_length]
        # If x has both a variable and a sample dimension, then add a singleton dimension to marginal_params to have the correct shape
        if marginal_params.dim() == x.dim():
            marginal_params = marginal_params[:, :, None, :]

        return self.marginal_flow.forward_no_logdet(marginal_params, x)

    def inverse(
        self,
        context: torch.Tensor,
        u: torch.Tensor,
        max_iter: int = 100,
        precision: float = 1e-6,
        max_value: float = 1000.0,
    ) -> torch.Tensor:
        """
        Compute the inverse cumulative density function of a marginal conditioned using the given context, for the given value of u.
        This method uses a dichotomic search.
        The gradient of this method cannot be computed, so it should only be used for sampling.

        Parameters:
        -----------
        context: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            The series and time steps dimensions are merged.
        u: Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            A tensor containing the value to be transformed using the inverse CDF.
            The series and time steps dimensions are merged.
            If a third dimension is present, then the context is considered to be constant across this dimension.
        max_iter: int, default = 100
            The maximum number of iterations for the dichotomic search.
            The precision of the result should improve by a factor of 2 at each iteration.
        precision: float, default = 1e-6
            If the difference between CDF(x) and u is less than this value for all variables, stop the search.
        max_value: float, default = 1000.0
            The absolute upper bound on the possible output.
        Returns:
        --------
        x: torch.Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            The inverse CDF at the given value.
            The series and time steps dimensions are merged.
            The shape of the output is the same as the shape of u.
        """
        marginal_params = self.marginal_conditioner(context)
        # If u has both a variable and a sample dimension, then add a singleton dimension to marginal_params to have the correct shape
        if marginal_params.dim() == u.dim():
            marginal_params = marginal_params[:, :, None, :]

        left = -max_value * torch.ones_like(u)
        right = max_value * torch.ones_like(u)
        for _ in range(max_iter):
            mid = (left + right) / 2
            error = self.marginal_flow.forward_no_logdet(marginal_params, mid) - u
            left[error <= 0] = mid[error <= 0]
            right[error >= 0] = mid[error >= 0]

            max_error = error.abs().max().item()
            if max_error < precision:
                break
        return mid



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from pathlib import Path
from typing import List, Tuple

import numpy as np
import pandas as pd
import logging
from gluonts.dataset.common import DataEntry, MetaData, Dataset, ListDataset
from gluonts.dataset.multivariate_grouper import MultivariateGrouper
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.repository._tsf_datasets import Dataset as MonashDataset
from gluonts.dataset.repository._tsf_datasets import datasets as monash_datasets
from gluonts.dataset.repository.datasets import (
    dataset_recipes,
    default_dataset_path,
    generate_forecasting_dataset,
    get_dataset,
    partial,
)


_DATA_BACKTEST_DEF = {
    "solar_10min": {  # 137 series, 72 prediction length
        "train_dates": [
            pd.Timestamp(year=2006, month=11, day=20),  # Monday
            pd.Timestamp(year=2006, month=11, day=27),
            pd.Timestamp(year=2006, month=12, day=4),
            pd.Timestamp(year=2006, month=12, day=11),
            pd.Timestamp(year=2006, month=12, day=18),
            pd.Timestamp(year=2006, month=12, day=25),
        ],
        "end_date": pd.Timestamp(year=2007, month=1, day=1),  # A Monday
    },
    "electricity_hourly": {  # 321 series, 24 prediction length
        "train_dates": [
            pd.Timestamp(year=2014, month=11, day=17),  # A Monday
            pd.Timestamp(year=2014, month=11, day=24),
            pd.Timestamp(year=2014, month=12, day=1),
            pd.Timestamp(year=2014, month=12, day=8),
            pd.Timestamp(year=2014, month=12, day=15),
            pd.Timestamp(year=2014, month=12, day=22),
        ],
        "end_date": pd.Timestamp(year=2014, month=12, day=29),  # Last Monday before end of data
    },
    "kdd_cup_2018_without_missing": {  # 270 series, 48 prediction length
        "train_dates": [
            pd.Timestamp(year=2018, month=1, day=1),  # A Monday
            pd.Timestamp(year=2018, month=1, day=15),
            pd.Timestamp(year=2018, month=1, day=29),
            pd.Timestamp(year=2018, month=2, day=12),
            pd.Timestamp(year=2018, month=2, day=26),
            pd.Timestamp(year=2018, month=3, day=12),
        ],
        "end_date": pd.Timestamp(year=2018, month=3, day=26),  # Last Monday before end of data
    },
    "traffic": {  # 862 series, 24 prediction length
        "train_dates": [
            pd.Timestamp(year=2016, month=11, day=14),  # A Monday
            pd.Timestamp(year=2016, month=11, day=21),
            pd.Timestamp(year=2016, month=11, day=28),
            pd.Timestamp(year=2016, month=12, day=5),
            pd.Timestamp(year=2016, month=12, day=12),
            pd.Timestamp(year=2016, month=12, day=19),
        ],
        "end_date": pd.Timestamp(year=2016, month=12, day=26),  # Last Monday before end of data
    },
    "fred_md": {  # 107 series, 12 prediction length
        "train_dates": [
            pd.Timestamp(year=2013, month=1, day=30),
            pd.Timestamp(year=2014, month=1, day=30),
            pd.Timestamp(year=2015, month=1, day=30),
            pd.Timestamp(year=2016, month=1, day=30),
            pd.Timestamp(year=2017, month=1, day=30),
            pd.Timestamp(year=2018, month=1, day=30),
        ],
        "end_date": pd.Timestamp(year=2019, month=1, day=30),  # Last January before end of data
    },
}

_DATA_PREBACKTEST_DEF = {
    "solar_10min": {  # 137 series, 72 prediction length
        "train_dates": [
            pd.Timestamp(year=2006, month=11, day=13),  # Monday
            pd.Timestamp(year=2006, month=11, day=20),
            pd.Timestamp(year=2006, month=11, day=27),
            pd.Timestamp(year=2006, month=12, day=4),
            pd.Timestamp(year=2006, month=12, day=11),
            pd.Timestamp(year=2006, month=12, day=18),
        ],
        "end_date": pd.Timestamp(year=2006, month=12, day=25),  # A Monday
    },
    "kdd_cup_2018_without_missing": {  # 270 series, 48 prediction length
        "train_dates": [
            pd.Timestamp(year=2017, month=12, day=18),  # A Monday
            pd.Timestamp(year=2018, month=1, day=1),
            pd.Timestamp(year=2018, month=1, day=15),
            pd.Timestamp(year=2018, month=1, day=29),
            pd.Timestamp(year=2018, month=2, day=12),
            pd.Timestamp(year=2018, month=2, day=26),
        ],
        "end_date": pd.Timestamp(year=2018, month=3, day=12),  # Last Monday before end of data
    },
    "fred_md": {  # 107 series, 12 prediction length
        "train_dates": [
            pd.Timestamp(year=2012, month=1, day=30),
            pd.Timestamp(year=2013, month=1, day=30),
            pd.Timestamp(year=2014, month=1, day=30),
            pd.Timestamp(year=2015, month=1, day=30),
            pd.Timestamp(year=2016, month=1, day=30),
            pd.Timestamp(year=2017, month=1, day=30),
        ],
        "end_date": pd.Timestamp(year=2018, month=1, day=30),  # Last January before end of data
    },
    "electricity_hourly": {  # 107 series, 12 prediction length
        "train_dates": [
            pd.Timestamp(year=2014, month=11, day=10),  # A Monday
            pd.Timestamp(year=2014, month=11, day=17),  # A Monday
            pd.Timestamp(year=2014, month=11, day=24),
            pd.Timestamp(year=2014, month=12, day=1),
            pd.Timestamp(year=2014, month=12, day=8),
            pd.Timestamp(year=2014, month=12, day=15),
        ],
        "end_date": pd.Timestamp(year=2014, month=12, day=22),  # Last Monday before end of data
    },
    "traffic": {  # 862 series, 24 prediction length
        "train_dates": [
            pd.Timestamp(year=2016, month=11, day=7),  # A Monday
            pd.Timestamp(year=2016, month=11, day=14),  # A Monday
            pd.Timestamp(year=2016, month=11, day=21),
            pd.Timestamp(year=2016, month=11, day=28),
            pd.Timestamp(year=2016, month=12, day=5),
            pd.Timestamp(year=2016, month=12, day=12),
        ],
        "end_date": pd.Timestamp(year=2016, month=12, day=19),  # Last Monday before end of data
    },
}


def _monash_inject_datasets(name, filename, record, prediction_length=None):
    """
    Injects datasets from the Monash Time Series Repository that were not included in GluonTS.
    """
    dataset_recipes.update(
        {
            name: partial(
                generate_forecasting_dataset,
                dataset_name=name,
                prediction_length=prediction_length,
            )
        }
    )
    monash_datasets.update({name: MonashDataset(file_name=filename, record=record)})


# Modifications to the GluonTS dataset repository
# * We add missing datasets from the Monash Time Series Repository
# * We rename datasets to have a prefix that makes their source explicit
_monash_inject_datasets(
    "electricity_hourly",
    "electricity_hourly_dataset.zip",
    "4656140",
    prediction_length=24,
)
_monash_inject_datasets("solar_10min", "solar_10_minutes_dataset.zip", "4656144", prediction_length=72)
_monash_inject_datasets("traffic", "traffic_hourly_dataset.zip", "4656132", prediction_length=24)


def _count_timesteps(left: pd.Timestamp, right: pd.Timestamp, delta: pd.DateOffset) -> int:
    """
    Count how many timesteps there are between left and right, according to the given timesteps delta.
    If the number if not integer, round down.
    """

    # This is due to GluonTS replacing Timestamp by Period for version 0.10.0.
    # Original code was tested on version 0.9.4
    if type(left) == pd.Period:
        left = left.to_timestamp()
    if type(right) == pd.Period:
        right = right.to_timestamp()

    assert right >= left, f"Case where left ({left}) is after right ({right}) is not implemented in _count_timesteps()."
    try:
        return (right - left) // delta
    except TypeError:
        # For MonthEnd offsets, the division does not work, so we count months one by one.
        for i in range(10000):
            if left + (i + 1) * delta > right:
                return i
        else:
            raise RuntimeError(
                f"Too large difference between both timestamps ({left} and {right}) for _count_timesteps()."
            )


def _load_raw_dataset(name: str, use_cached: bool) -> Tuple[MetaData, List[DataEntry]]:
    """
    Load the dataset using GluonTS method, and combining both the train and test data.

    The combination is needed due to going through GluonTS, and could be avoided by loading the data directly from
    the Monash repository.
    """
    # Where to save the downloaded input files
    cache_path = Path(os.environ.get("TACTIS_DATA_STORE", default_dataset_path))
    # This gives the univariate version of the dataset, with pre-made train/test split
    uv_dataset = get_dataset(name, regenerate=not use_cached, path=cache_path)

    # Combine the training and testing portion of the dataset, giving a single dataset containing all the data
    timestep_delta = pd.tseries.frequencies.to_offset(uv_dataset.metadata.freq)

    data = [series.copy() for series in uv_dataset.train]

    # This would be more efficient if done in reverse order, but we only have an iterator, not a list
    for i, new_series in enumerate(uv_dataset.test):
        # The test datasets are in the same order as the train datasets, but repeated for each test timestamp
        old_series = data[i % len(data)]

        # Make sure we don't combine series that we know are distinct
        if "feat_static_cat" in new_series:
            assert old_series["feat_static_cat"] == new_series["feat_static_cat"]

        if old_series["start"] > new_series["start"]:
            extra_timesteps = _count_timesteps(new_series["start"], old_series["start"], timestep_delta)
            # Not robust, but at the very least check that first common entry is the same before combining the data
            assert old_series["target"][0] == new_series["target"][extra_timesteps]
            old_series["start"] = new_series["start"]
            old_series["target"] = np.concatenate([new_series["target"][0:extra_timesteps], old_series["target"]])

        old_end = old_series["start"] + len(old_series["target"]) * timestep_delta
        new_end = new_series["start"] + len(new_series["target"]) * timestep_delta
        if new_end > old_end:
            extra_timesteps = _count_timesteps(old_end, new_end, timestep_delta)
            assert old_series["target"][-1] == new_series["target"][-extra_timesteps - 1]
            old_series["target"] = np.concatenate([old_series["target"], new_series["target"][-extra_timesteps:]])

    return uv_dataset.metadata, data


def generate_hp_search_datasets(
    name: str, history_length_multiple: float, use_cached: bool = True
) -> Tuple[MetaData, Dataset, Dataset]:
    """
    Generate the training and validation datasets to be used during the hyperparameter search.

    The validation dataset always ends at the timestep of the first backtesting period.
    The length of the validation period is equal to 7 times the prediction length, plus the needed history length.
    The training dataset ends at the beginning of the validation dataset (ignoring the needed history length),
    and starts at the beginning of the full dataset.

    Parameters:
    -----------
    name: str
        The name of the dataset.
    history_length_multiple: float
        The length of the history that will be sent to the model, as a multiple of the dataset prediction length.
        The result is rounded down to the nearest integer.
    use_cached: bool, default to True
        If set to True, use the cached version of the data if available.

    Returns
    -------
    metadata: MetaData
        The MetaData of the dataset.
    train_data: Dataset
        The training dataset.
    valid_data: Dataset
        The validation dataset.
    """
    metadata, raw_dataset = _load_raw_dataset(name, use_cached=use_cached)

    first_backtest_timestamp = _DATA_BACKTEST_DEF[name]["train_dates"][0]
    history_length = int(history_length_multiple * metadata.prediction_length)
    validation_length = 7 * metadata.prediction_length

    timestep_delta = pd.tseries.frequencies.to_offset(metadata.freq)

    train_data = []
    valid_data = []
    for i, series in enumerate(raw_dataset):
        first_backtest_index = _count_timesteps(
            series["start"], first_backtest_timestamp, timestep_delta
        )  # 647 for FRED

        # train_start_index is always 0
        # train_end_index is always the start of the validation dataset
        train_end_index = first_backtest_index - validation_length  # 563 for FRED

        # validation set starts 7 `frequencies` after the training set
        # validation_start_index is always `history_length` steps behind the start of the validation set
        # validation_end_index is always at the start of the backtest set
        validation_start_index = first_backtest_index - validation_length - history_length  # 539 for FRED

        s_train = series.copy()
        s_train["target"] = series["target"][
            :train_end_index
        ]  # series["target"] is 728, after slicing s_train["target"] is 563
        s_train["item_id"] = i
        train_data.append(s_train)

        s_valid = series.copy()
        s_valid["start"] = (
            s_valid["start"] + validation_length * timestep_delta
        )  # Before slicing: 1959-01-31 00:00:00, after slicing: 1966-01-31 00:00:00
        s_valid["target"] = series["target"][
            validation_start_index:first_backtest_index
        ]  # series["target"] is 728, after slicing 539:647, it is of length 108
        s_valid["item_id"] = i
        valid_data.append(s_valid)

    # MultivariateGrouper call operation is not without side-effect, so we need two independant ones.
    train_grouper = MultivariateGrouper()
    valid_grouper = MultivariateGrouper()

    return metadata, train_grouper(train_data), valid_grouper(valid_data)


def maximum_backtest_id(name: str) -> int:
    """
    Return the largest possible backtesting id for the given dataset.

    Parameters:
    -----------
    name: str
        The name of the dataset.

    Returns
    -------
    maximum_id
        The largest value for the backtest_id parameter in generate_backtesting_datasets().
    """
    return len(_DATA_BACKTEST_DEF[name]["train_dates"])


class __FixedMultivariateGrouper(MultivariateGrouper):
    """
    Temporary fix for MultivariateGrouper when used with NumPy >= 1.24.
    See: https://github.com/awslabs/gluonts/issues/2612
    """

    def _prepare_test_data(self, dataset: Dataset) -> Dataset:
        assert self.num_test_dates is not None
        assert len(dataset) % self.num_test_dates == 0

        logging.info("group test time series to datasets")

        test_length = len(dataset) // self.num_test_dates

        all_entries = list()
        for test_start in range(0, len(dataset), test_length):
            dataset_at_test_date = dataset[test_start : test_start + test_length]
            transformed_target = self._transform_target(self._left_pad_data, dataset_at_test_date)[FieldName.TARGET]

            grouped_data = dict()
            grouped_data[FieldName.TARGET] = np.array(list(transformed_target), dtype=np.float32)
            for data in dataset:
                fields = data.keys()
                break
            if FieldName.FEAT_DYNAMIC_REAL in fields:
                grouped_data[FieldName.FEAT_DYNAMIC_REAL] = np.vstack(
                    [data[FieldName.FEAT_DYNAMIC_REAL] for data in dataset],
                )
            grouped_data = self._restrict_max_dimensionality(grouped_data)
            grouped_data[FieldName.START] = self.first_timestamp
            grouped_data[FieldName.FEAT_STATIC_CAT] = [0]
            all_entries.append(grouped_data)

        return ListDataset(all_entries, freq=self.frequency, one_dim_target=False)


def generate_prebacktesting_datasets(
    name: str, backtest_id: int, history_length_multiple: float, use_cached: bool = True
) -> Tuple[MetaData, Dataset, Dataset]:
    """
    Generate the training and testing datasets to be used during the backtesting.

    The training dataset ends at the timestamp associated with the given backtesting id.
    The testing dataset contains multiple testing instances, each separated by the prediction length,
    starting from its backtesting timestamp to the next backtesting period timestamp (or the ending timestamp).

    Parameters:
    -----------
    name: str
        The name of the dataset.
    backtest_id: int
        The identifier for the backtesting period. Its maximum value can be told by maximum_backtest_id().
    history_length_multiple: float
        The length of the history that will be sent to the model, as a multiple of the dataset prediction length.
        The result is rounded down to the nearest integer.
    use_cached: bool, default to True
        If set to True, use the cached version of the data if available.

    Returns
    -------
    metadata: MetaData
        The MetaData of the dataset.
    train_data: Dataset
        The training dataset.
    test_data: Dataset
        The testing dataset.
    """
    metadata, raw_dataset = _load_raw_dataset(name, use_cached=use_cached)

    backtest_timestamp = _DATA_PREBACKTEST_DEF[name]["train_dates"][backtest_id]
    history_length = int(history_length_multiple * metadata.prediction_length)

    timestep_delta = pd.tseries.frequencies.to_offset(metadata.freq)
    test_offset = timestep_delta * metadata.prediction_length
    if backtest_id + 1 < maximum_backtest_id(name):
        num_test_dates = _count_timesteps(
            _DATA_PREBACKTEST_DEF[name]["train_dates"][backtest_id],
            _DATA_PREBACKTEST_DEF[name]["train_dates"][backtest_id + 1],
            test_offset,
        )
    else:
        num_test_dates = _count_timesteps(
            _DATA_PREBACKTEST_DEF[name]["train_dates"][backtest_id],
            _DATA_PREBACKTEST_DEF[name]["end_date"],
            test_offset,
        )

    train_data = []
    for i, series in enumerate(raw_dataset):
        train_end_index = _count_timesteps(series["start"], backtest_timestamp, timestep_delta)

        s_train = series.copy()
        s_train["target"] = series["target"][:train_end_index]
        s_train["item_id"] = i
        train_data.append(s_train)

    # GluonTS multivariate format for multiple tests is ordered first by date, then by series.
    test_data = []
    for test_id in range(num_test_dates):
        for i, series in enumerate(raw_dataset):
            train_end_index = _count_timesteps(series["start"], backtest_timestamp, timestep_delta)
            test_end_index = train_end_index + metadata.prediction_length * (test_id + 1)
            test_start_index = test_end_index - metadata.prediction_length - history_length

            s_test = series.copy()
            s_test["start"] = series["start"] + test_start_index * timestep_delta
            s_test["target"] = series["target"][test_start_index:test_end_index]
            s_test["item_id"] = len(test_data)
            test_data.append(s_test)

    train_grouper = MultivariateGrouper()
    test_grouper = __FixedMultivariateGrouper(num_test_dates=num_test_dates)

    return metadata, train_grouper(train_data), test_grouper(test_data)


def generate_backtesting_datasets(
    name: str, backtest_id: int, history_length_multiple: float, use_cached: bool = True
) -> Tuple[MetaData, Dataset, Dataset]:
    """
    Generate the training and testing datasets to be used during the backtesting.

    The training dataset ends at the timestamp associated with the given backtesting id.
    The testing dataset contains multiple testing instances, each separated by the prediction length,
    starting from its backtesting timestamp to the next backtesting period timestamp (or the ending timestamp).

    Parameters:
    -----------
    name: str
        The name of the dataset.
    backtest_id: int
        The identifier for the backtesting period. Its maximum value can be told by maximum_backtest_id().
    history_length_multiple: float
        The length of the history that will be sent to the model, as a multiple of the dataset prediction length.
        The result is rounded down to the nearest integer.
    use_cached: bool, default to True
        If set to True, use the cached version of the data if available.

    Returns
    -------
    metadata: MetaData
        The MetaData of the dataset.
    train_data: Dataset
        The training dataset.
    test_data: Dataset
        The testing dataset.
    """
    metadata, raw_dataset = _load_raw_dataset(name, use_cached=use_cached)

    backtest_timestamp = _DATA_BACKTEST_DEF[name]["train_dates"][backtest_id]
    history_length = int(history_length_multiple * metadata.prediction_length)

    timestep_delta = pd.tseries.frequencies.to_offset(metadata.freq)
    test_offset = timestep_delta * metadata.prediction_length
    if backtest_id + 1 < maximum_backtest_id(name):
        num_test_dates = _count_timesteps(
            _DATA_BACKTEST_DEF[name]["train_dates"][backtest_id],
            _DATA_BACKTEST_DEF[name]["train_dates"][backtest_id + 1],
            test_offset,
        )
    else:
        num_test_dates = _count_timesteps(
            _DATA_BACKTEST_DEF[name]["train_dates"][backtest_id],
            _DATA_BACKTEST_DEF[name]["end_date"],
            test_offset,
        )

    train_data = []
    for i, series in enumerate(raw_dataset):
        train_end_index = _count_timesteps(series["start"], backtest_timestamp, timestep_delta)

        s_train = series.copy()
        s_train["target"] = series["target"][:train_end_index]
        s_train["item_id"] = i
        train_data.append(s_train)

    # GluonTS multivariate format for multiple tests is ordered first by date, then by series.
    test_data = []
    for test_id in range(num_test_dates):
        for i, series in enumerate(raw_dataset):
            train_end_index = _count_timesteps(series["start"], backtest_timestamp, timestep_delta)
            test_end_index = train_end_index + metadata.prediction_length * (test_id + 1)
            test_start_index = test_end_index - metadata.prediction_length - history_length

            s_test = series.copy()
            s_test["start"] = series["start"] + test_start_index * timestep_delta
            s_test["target"] = series["target"][test_start_index:test_end_index]
            s_test["item_id"] = len(test_data)
            test_data.append(s_test)

    train_grouper = MultivariateGrouper()
    test_grouper = __FixedMultivariateGrouper(num_test_dates=num_test_dates)

    return metadata, train_grouper(train_data), test_grouper(test_data)



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import torch
import time
from torch import nn


class Encoder(nn.Module):
    """
    The traditional encoder for TACTiS, based on the Transformer architecture.

    The encoder receives an input which contains for each series and time step:
    * The series value at the time step, masked to zero if part of the values to be forecasted
    * The mask
    * The embedding for the series
    * The embedding for the time step
    And has already been through any input encoder.

    The decoder returns an output containing an embedding for each series and time step.
    """

    def __init__(
        self,
        attention_layers: int,
        attention_heads: int,
        attention_dim: int,
        attention_feedforward_dim: int,
        dropout: float = 0.1,
    ):
        """
        Parameters:
        -----------
        attention_layers: int
            How many successive attention layers this encoder will use.
        attention_heads: int
            How many independant heads the attention layer will have.
        attention_dim: int
            The size of the attention layer input and output, for each head.
        attention_feedforward_dim: int
            The dimension of the hidden layer in the feed forward step.
        dropout: float, default to 0.1
            Dropout parameter for the attention.
        """
        super().__init__()

        self.attention_layers = attention_layers
        self.attention_heads = attention_heads
        self.attention_dim = attention_dim
        self.attention_feedforward_dim = attention_feedforward_dim
        self.dropout = dropout

        self.transformer_encoder = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(
                self.attention_dim * self.attention_heads,
                self.attention_heads,
                self.attention_feedforward_dim,
                self.dropout,
            ),
            self.attention_layers,
        )
        self.total_attention_time = 0.0

    @property
    def embedding_dim(self) -> int:
        """
        Returns:
        --------
        dim: int
            The expected dimensionality of the input embedding, and the dimensionality of the output embedding
        """
        return self.attention_dim * self.attention_heads

    def forward(self, encoded: torch.Tensor) -> torch.Tensor:
        """
        Compute the embedding for each series and time step.

        Parameters:
        -----------
        encoded: Tensor [batch, series, time steps, input embedding dimension]
            A tensor containing an embedding for each series and time step.
            This embedding is expected to only contain local information, with no interaction between series or time steps.

        Returns:
        --------
        output: torch.Tensor [batch, series, time steps, output embedding dimension]
            The transformed embedding for each series and time step.
        """
        num_batches = encoded.shape[0]
        num_series = encoded.shape[1]
        num_timesteps = encoded.shape[2]

        # Merge the series and time steps, since the PyTorch attention implementation only accept three-dimensional input,
        # and the attention is applied between all tokens, no matter their series or time step.
        encoded = encoded.view(num_batches, num_series * num_timesteps, self.embedding_dim)

        attention_start_time = time.time()

        # The PyTorch implementation wants the following order: [tokens, batch, embedding]
        encoded = encoded.transpose(0, 1)
        # Arjun: I have changed this to the usual full attention
        output = self.transformer_encoder(encoded)
        # # This was the previous code
        # output = self.transformer_encoder(
        #     encoded, mask=torch.zeros(encoded.shape[0], encoded.shape[0], device=encoded.device)
        # )
        # Reset to the original shape
        output = output.transpose(0, 1)
        attention_end_time = time.time()
        self.total_attention_time = attention_end_time - attention_start_time

        # Resize back to original shape
        output = output.view(num_batches, num_series, num_timesteps, self.embedding_dim)

        return output


class TemporalEncoder(nn.Module):
    """
    The encoder for TACTiS, based on the Temporal Transformer architecture.
    This encoder alternate between doing self-attention between different series of the same time steps,
    and doing self-attention between different time steps of the same series.
    This greatly reduces the memory footprint compared to TACTiSEncoder.

    The encoder receives an input which contains for each variable and time step:
    * The series value at the time step, masked to zero if part of the values to be forecasted
    * The mask
    * The embedding for the series
    * The embedding for the time step
    And has already been through any input encoder.

    The decoder returns an output containing an embedding for each series and time step.
    """

    def __init__(
        self,
        attention_layers: int,
        attention_heads: int,
        attention_dim: int,
        attention_feedforward_dim: int,
        dropout: float = 0.1,
    ):
        """
        Parameters:
        -----------
        attention_layers: int
            How many successive attention pairs of layers this will use.
            Note that the total number of layers is going to be the double of this number.
            Each pair will consist of a layer with attention done over time steps,
            followed by a layer with attention done over series.
        attention_heads: int
            How many independant heads the attention layer will have.
        attention_dim: int
            The size of the attention layer input and output, for each head.
        attention_feedforward_dim: int
            The dimension of the hidden layer in the feed forward step.
        dropout: float, default to 0.1
            Dropout parameter for the attention.
        """
        super().__init__()

        self.attention_layers = attention_layers
        self.attention_heads = attention_heads
        self.attention_dim = attention_dim
        self.attention_feedforward_dim = attention_feedforward_dim
        self.dropout = dropout
        self.total_attention_time = 0.0

        self.layer_timesteps = nn.ModuleList(
            [
                nn.TransformerEncoderLayer(
                    self.attention_dim * self.attention_heads,
                    self.attention_heads,
                    self.attention_feedforward_dim,
                    self.dropout,
                )
                for _ in range(self.attention_layers)
            ]
        )

        self.layer_series = nn.ModuleList(
            [
                nn.TransformerEncoderLayer(
                    self.attention_dim * self.attention_heads,
                    self.attention_heads,
                    self.attention_feedforward_dim,
                    self.dropout,
                )
                for _ in range(self.attention_layers)
            ]
        )

    @property
    def embedding_dim(self) -> int:
        """
        Returns:
        --------
        dim: int
            The expected dimensionality of the input embedding, and the dimensionality of the output embedding
        """
        return self.attention_dim * self.attention_heads

    def forward(self, encoded: torch.Tensor) -> torch.Tensor:
        """
        Compute the embedding for each series and time step.

        Parameters:
        -----------
        encoded: Tensor [batch, series, time steps, input embedding dimension]
            A tensor containing an embedding for each series and time step.
            This embedding is expected to only contain local information, with no interaction between series or time steps.

        Returns:
        --------
        output: torch.Tensor [batch, series, time steps, output embedding dimension]
            The transformed embedding for each series and time step.
        """
        num_batches = encoded.shape[0]
        num_series = encoded.shape[1]
        num_timesteps = encoded.shape[2]

        data = encoded

        attention_start_time = time.time()
        for i in range(self.attention_layers):
            # Treat the various series as a batch dimension
            mod_timesteps = self.layer_timesteps[i]
            # [batch * series, time steps, embedding]
            data = data.flatten(start_dim=0, end_dim=1)
            # [time steps, batch * series, embedding] Correct order for PyTorch module
            data = data.transpose(0, 1)
            # Perform attention
            data = mod_timesteps(data)
            # [batch * series, time steps, embedding]
            data = data.transpose(0, 1)
            # [batch, series, time steps, embedding]
            data = data.unflatten(dim=0, sizes=(num_batches, num_series))

            # Treat the various time steps as a batch dimension
            mod_series = self.layer_series[i]
            # Transpose to [batch, timesteps, series, embedding]
            data = data.transpose(1, 2)
            # [batch * time steps, series, embedding] Correct order for PyTorch module
            data = data.flatten(start_dim=0, end_dim=1)
            # [series, batch * time steps, embedding]
            data = data.transpose(0, 1)
            # Perform attention
            data = mod_series(data)
            # [batch * time steps, series, embedding]
            data = data.transpose(0, 1)
            # [batch, time steps, series, embedding]
            data = data.unflatten(dim=0, sizes=(num_batches, num_timesteps))
            # Transpose to [batch, series, time steps, embedding]
            data = data.transpose(1, 2)

        attention_end_time = time.time()
        self.total_attention_time = attention_end_time - attention_start_time
        # The resulting tensor may not be contiguous, which can cause problems further down the line.
        output = data.contiguous()

        return output



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict

import torch
from torch import nn

from ..model.tactis import TACTiS


class TACTiSTrainingNetwork(nn.Module):
    """
    A shell on top of the TACTiS module, to be used during training only.
    """

    def __init__(
        self,
        num_series: int,
        model_parameters: Dict[str, Any],
    ):
        """
        Parameters:
        -----------
        num_series: int
            Number of series of the data which will be sent to the model.
        model_parameters: Dict[str, Any]
            The parameters of the underlying TACTiS model, as a dictionary.
        """
        super().__init__()

        self.model = TACTiS(num_series, **model_parameters)

    def forward(
        self,
        past_target_norm: torch.Tensor,
        future_target_norm: torch.Tensor,
    ) -> torch.Tensor:
        """
        Parameters:
        -----------
        past_target_norm: torch.Tensor [batch, time steps, series]
            The historical data that will be available at inference time.
        future_target_norm: torch.Tensor [batch, time steps, series]
            The data to be forecasted at inference time.

        Returns:
        --------
        loss: torch.Tensor []
            The loss function, averaged over all batches.
        """
        # The data coming from Gluon is not in the shape we use in the model, so transpose it.
        hist_value = past_target_norm.transpose(1, 2)
        pred_value = future_target_norm.transpose(1, 2)

        # For the time steps, we take for granted that the data is aligned with a constant frequency
        hist_time = torch.arange(0, hist_value.shape[2], dtype=int, device=hist_value.device)[None, :].expand(
            hist_value.shape[0], -1
        )
        pred_time = torch.arange(
            hist_value.shape[2],
            hist_value.shape[2] + pred_value.shape[2],
            dtype=int,
            device=pred_value.device,
        )[None, :].expand(pred_value.shape[0], -1)

        return self.model.loss(
            hist_time=hist_time,
            hist_value=hist_value,
            pred_time=pred_time,
            pred_value=pred_value,
        )


class TACTiSPredictionNetwork(nn.Module):
    """
    A shell on top of the TACTiS module, to be used during inference only.
    """

    def __init__(
        self,
        num_series: int,
        model_parameters: Dict[str, Any],
        prediction_length: int,
        num_parallel_samples: int,
    ):
        """
        Parameters:
        -----------
        num_series: int
            Number of series of the data which will be sent to the model.
        model_parameters: Dict[str, Any]
            The parameters of the underlying TACTiS model, as a dictionary.
        """
        super().__init__()

        self.model = TACTiS(num_series, **model_parameters)
        self.num_parallel_samples = num_parallel_samples
        self.prediction_length = prediction_length

    def forward(self, past_target_norm: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
        -----------
        past_target_norm: torch.Tensor [batch, time steps, series]
            The historical data that are available.

        Returns:
        --------
        samples: torch.Tensor [samples, batch, time steps, series]
            Samples from the forecasted distribution.
        """
        # The data coming from Gluon is not in the shape we use in the model, so transpose it.
        hist_value = past_target_norm.transpose(1, 2)

        # For the time steps, we take for granted that the data is aligned with a constant frequency
        hist_time = torch.arange(0, hist_value.shape[2], dtype=int, device=hist_value.device)[None, :].expand(
            hist_value.shape[0], -1
        )
        pred_time = torch.arange(
            hist_value.shape[2],
            hist_value.shape[2] + self.prediction_length,
            dtype=int,
            device=hist_value.device,
        )[None, :].expand(hist_value.shape[0], -1)

        samples = self.model.sample(
            num_samples=self.num_parallel_samples,
            hist_time=hist_time,
            hist_value=hist_value,
            pred_time=pred_time,
        )

        # The model decoder returns both the observed and sampled values, so removed the observed ones.
        # Also, reorder from [batch, series, time steps, samples] to GluonTS expected [batch, samples, time steps, series].
        return samples[:, :, -self.prediction_length :, :].permute((0, 3, 2, 1))


class TACTiSPredictionNetworkInterpolation(nn.Module):
    """
    A shell on top of the TACTiS module, to be used during inference only.
    For now, interpolation is only supported with equal history before and after the window to be interpolated.
    """

    def __init__(
        self,
        num_series: int,
        model_parameters: Dict[str, Any],
        prediction_length: int,
        history_length: int,
        num_parallel_samples: int,
    ):
        """
        Parameters:
        -----------
        num_series: int
            Number of series of the data which will be sent to the model.
        model_parameters: Dict[str, Any]
            The parameters of the underlying TACTiS model, as a dictionary.
        """
        super().__init__()

        self.model = TACTiS(num_series, **model_parameters)
        self.num_parallel_samples = num_parallel_samples
        self.prediction_length = prediction_length
        self.history_length = history_length

    def forward(self, past_target_norm: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
        -----------
        past_target_norm: torch.Tensor [batch, time steps, series]
            The historical data that are available.

        Returns:
        --------
        samples: torch.Tensor [samples, batch, time steps, series]
            Samples from the forecasted distribution.
        """
        # The data coming from Gluon is not in the shape we use in the model, so transpose it.
        hist_value = past_target_norm.transpose(1, 2)

        # For the time steps, we take for granted that the data is aligned with a constant frequency
        hist_time = torch.arange(0, hist_value.shape[2], dtype=int, device=hist_value.device)[None, :].expand(
            hist_value.shape[0], -1
        )
        pred_time = torch.arange(
            hist_value.shape[2],
            hist_value.shape[2] + self.prediction_length,
            dtype=int,
            device=hist_value.device,
        )[None, :].expand(hist_value.shape[0], -1)

        samples = self.model.sample(
            num_samples=self.num_parallel_samples,
            hist_time=hist_time,
            hist_value=hist_value,
            pred_time=pred_time,
        )

        total_length = samples.shape[2]
        interpolation_window_start = total_length - self.prediction_length - (self.history_length // 2)
        interpolation_window_end = total_length - (self.history_length // 2)

        # The model decoder returns both the observed and sampled values, so removed the observed ones.
        # Also, reorder from [batch, series, time steps, samples] to GluonTS expected [batch, samples, time steps, series].
        return samples[:, :, interpolation_window_start:interpolation_window_end, :].permute((0, 3, 2, 1))



# MIT License

# Copyright (c) 2019 Chin-Wei Huang

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Adapted from: https://github.com/CW-Huang/torchkit/blob/master/torchkit/flows.py

import math
import torch
from torch import nn

# To be compared to the float32 machine epsilon of 2**-23 ~= 1.2e-7
EPSILON = 1e-6


def log_sum_exp(A, dim=-1, keepdim=False):
    """
    Compute a sum in logarithm space: log(exp(a) + exp(b))
    Properly handle values which exponential cannot be represented in float32.
    """
    max_A = A.max(axis=dim, keepdim=True).values
    norm_A = A - max_A
    result = torch.log(torch.exp(norm_A).sum(axis=dim, keepdim=True)) + max_A
    if keepdim:
        return result
    else:
        return torch.squeeze(result, dim=dim)


def log_sigmoid(x):
    """
    Logarithm of the sigmoid function.
    Substract the epsilon to avoid 0 as a possible value for large x.
    """
    return -nn.functional.softplus(-x) - EPSILON


def act_a(x):
    return nn.functional.softplus(x) + EPSILON


def act_b(x):
    return x


def act_w(x):
    return nn.functional.softmax(x, dim=-1)


class SigmoidFlow(nn.Module):
    # Indices:
    # b: batch
    # s: sample
    # o: output dimension
    # h: hidden dimension
    # i: input dimension

    def __init__(self, hidden_dim: int, no_logit: bool = False):
        """
        A single layer of the Deep Sigmoid Flow network.
        Does not contains its parameters, they must be sent in the forward method.

        Parameters:
        -----------
        hidden_dim: uint
            The number of hidden units
        no_logit: bool, default to False
            If set to True, then the network will return a value in the (0, 1) interval.
            If kept to False, then the network will apply a logit function to this value to return
            a value in the (-inf, inf) interval.
        """
        super(SigmoidFlow, self).__init__()
        self.hidden_dim = hidden_dim
        self.no_logit = no_logit

        self.act_a = act_a
        self.act_b = act_b
        self.act_w = act_w

    def forward(self, params, x, logdet):
        """
        Transform the given value according to the given parameters,
        computing the derivative of the transformation at the same time.
        params third dimension must be equal to 3 times the number of hidden units.
        """
        # Indices:
        # b: batch
        # v: variables
        # h: hidden dimension

        # params: b, v, 3*h
        # x: b, v
        # logdet: b
        # output x: b, v
        # output logdet: b
        assert params.shape[-1] == 3 * self.hidden_dim

        a = self.act_a(params[..., : self.hidden_dim])  # b, v, h
        b = self.act_b(params[..., self.hidden_dim : 2 * self.hidden_dim])  # b, v, h
        pre_w = params[..., 2 * self.hidden_dim :]  # b, v, h
        w = self.act_w(pre_w)  # b, v, h

        pre_sigm = a * x[..., None] + b  # b, v, h
        sigm = torch.sigmoid(pre_sigm)  # b, v, h
        x_pre = (w * sigm).sum(dim=-1)  # b, v

        logj = (
            nn.functional.log_softmax(pre_w, dim=-1) + log_sigmoid(pre_sigm) + log_sigmoid(-pre_sigm) + torch.log(a)
        )  # b, v, h

        logj = log_sum_exp(logj, dim=-1, keepdim=False)  # b, v

        if self.no_logit:
            # Only keep the batch dimension, summing all others in case this method is called with more dimensions
            # Adding the passed logdet here to accumulate
            logdet = logj.sum(dim=tuple(range(1, logj.dim()))) + logdet
            return x_pre, logdet

        x_pre_clipped = x_pre * (1 - EPSILON) + EPSILON * 0.5  # b, v
        xnew = torch.log(x_pre_clipped) - torch.log(1 - x_pre_clipped)  # b, v

        logdet_ = logj + math.log(1 - EPSILON) - (torch.log(x_pre_clipped) + torch.log(-x_pre_clipped + 1))  # b, v

        # Only keep the batch dimension, summing all others in case this method is called with more dimensions
        logdet = logdet_.sum(dim=tuple(range(1, logdet_.dim()))) + logdet

        return xnew, logdet

    def forward_no_logdet(self, params, x):
        """
        Transform the given value according to the given parameters,
        but does not compute the derivative of the transformation.
        params third dimension must be equal to 3 times the number of hidden units.
        """
        # Indices: (when used for inversion)
        # b: batch
        # s: samples
        # h: hidden dimension

        # params: b, s, 3*h
        # x: b, s
        # output x: b, s
        assert params.shape[-1] == 3 * self.hidden_dim

        a = self.act_a(params[..., : self.hidden_dim])  # b, s, h
        b = self.act_b(params[..., self.hidden_dim : 2 * self.hidden_dim])  # b, s, h
        pre_w = params[..., 2 * self.hidden_dim :]  # b, s, h
        w = self.act_w(pre_w)  # b, s, h

        pre_sigm = a * x[..., None] + b  # b, s, h
        sigm = torch.sigmoid(pre_sigm)  # b, s, h
        x_pre = (w * sigm).sum(dim=-1)  # b, s

        if self.no_logit:
            return x_pre

        x_pre_clipped = x_pre * (1 - EPSILON) + EPSILON * 0.5  # b, s
        xnew = torch.log(x_pre_clipped) - torch.log(1 - x_pre_clipped)  # b, s

        return xnew


class DeepSigmoidFlow(nn.Module):
    def __init__(self, n_layers: int, hidden_dim: int):
        """
        A Deep Sigmoid Flow network, made of multiple Sigmoig Flow layers.
        This model a flexible transformation from any real values into the (0, 1) interval.
        Does not contains its parameters, they must be sent in the forward method.

        Parameters:
        -----------
        n_layers: uint
            The number of sigmoid flow layers
        hidden_dim: uint
            The number of hidden units
        """
        super(DeepSigmoidFlow, self).__init__()

        self.params_length = 3 * hidden_dim

        elayers = nn.ModuleList([SigmoidFlow(hidden_dim) for _ in range(n_layers - 1)])
        elayers += [SigmoidFlow(hidden_dim, no_logit=True)]
        self.layers = nn.Sequential(*elayers)

    @property
    def total_params_length(self):
        return len(self.layers) * self.params_length

    def forward(self, params, x):
        """
        Transform the given value according to the given parameters,
        computing the derivative of the transformation at the same time.
        params third dimension must be equal to total_params_length.
        """
        # params: batches, variables, params dim
        # x: batches, variables
        logdet = torch.zeros(
            x.shape[0],
        ).to(x.device)
        for i, layer in enumerate(self.layers):
            x, logdet = layer(
                params[..., i * self.params_length : (i + 1) * self.params_length],
                x,
                logdet,
            )
        return x, logdet

    def forward_no_logdet(self, params, x):
        """
        Transform the given value according to the given parameters,
        but does not compute the derivative of the transformation.
        params third dimension must be equal to total_params_length.
        """
        # params: batches, samples, params dim
        # x: batches, samples
        for i, layer in enumerate(self.layers):
            x = layer.forward_no_logdet(params[..., i * self.params_length : (i + 1) * self.params_length], x)
        return x

    def inverse(
        self,
        marginal_params: torch.Tensor,
        u: torch.Tensor,
        max_iter: int = 100,
        precision: float = 1e-6,
        max_value: float = 1000.0,
    ) -> torch.Tensor:
        """

        NOTE: Added for compatability with Alex' notebook "ground_truth_copula.ipynb" demonstrating the 2-dimensional example
        This function actually belongs to the "marginal" class

        Compute the inverse cumulative density function of a marginal conditioned using the given context, for the given value of u.
        This method uses a dichotomic search.
        The gradient of this method cannot be computed, so it should only be used for sampling.

        Parameters:
        -----------
        context: Tensor [batch, series * time steps, embedding dimension]
            A tensor containing an embedding for each variable and time step.
            The series and time steps dimensions are merged.
        u: Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            A tensor containing the value to be transformed using the inverse CDF.
            The series and time steps dimensions are merged.
            If a third dimension is present, then the context is considered to be constant across this dimension.
        max_iter: int, default = 100
            The maximum number of iterations for the dichotomic search.
            The precision of the result should improve by a factor of 2 at each iteration.
        precision: float, default = 1e-6
            If the difference between CDF(x) and u is less than this value for all variables, stop the search.
        max_value: float, default = 1000.0
            The absolute upper bound on the possible output.
        Returns:
        --------
        x: torch.Tensor [batch, series * time steps] or [batch, series * time steps, samples]
            The inverse CDF at the given value.
            The series and time steps dimensions are merged.
            The shape of the output is the same as the shape of u.
        """
        # If u has both a variable and a sample dimension, then add a singleton dimension to marginal_params to have the correct shape
        if marginal_params.dim() == u.dim():
            marginal_params = marginal_params[:, :, None, :]

        left = -max_value * torch.ones_like(u)
        right = max_value * torch.ones_like(u)
        for _ in range(max_iter):
            mid = (left + right) / 2
            error = self.forward_no_logdet(marginal_params, mid) - u
            left[error <= 0] = mid[error <= 0]
            right[error >= 0] = mid[error >= 0]

            max_error = error.abs().max().item()
            if max_error < precision:
                break
        return mid
